from collections import defaultdict
from unittest.mock import patch, call, MagicMock

from bs4 import BeautifulSoup
import pytest

# from osd_slides.main import (
#     fefmalsmc,
#     showDownloadablePdf,
#     searchAndDownloadPdf,
#     downloadPdf,
#     readAvailableHtmlDocs,
# )
import sys
import os


from ..main import Downloader
from ..search import Search

downloader = Downloader()
url = "https://www.cs.columbia.edu/~paine/4995/lectures/"


@patch("builtins.print")
def test_main0(mock_print):
    testargs = ["prog", "download", "random-stuff", "more random"]
    with patch.object(sys, "argv", testargs):
        downloader.fefmalsmc()  # changed
        assert mock_print.call_args.args == ("Too many arguments",)


@patch("builtins.print")
def test_main1(mock_print):
    testargs = ["prog"]
    with patch.object(sys, "argv", testargs):
        downloader.fefmalsmc()
        assert mock_print.call_args.args == ("Too few arguments",)


# basically mock the function it is calling


def test_main2(mocker):
    mocked_searchAndDownloadPdf = mocker.patch("osd_slides.main.Downloader.searchAndDownloadPdf")  # noqa: E501

    testargs = ["prog", "download"]
    with patch.object(sys, "argv", testargs):
        downloader.fefmalsmc()
        mocked_searchAndDownloadPdf.assert_called_once()


def test_main3(mocker):
    mocked_searchAndDownloadPdf = mocker.patch("osd_slides.main.Downloader.showDownloadablePdf")  # noqa: E501
    testargs = ["prog", "show"]
    with patch.object(sys, "argv", testargs):
        downloader.fefmalsmc()
        mocked_searchAndDownloadPdf.assert_called_once_with()


# TODO one more case where nothing is called


def test_searchAndDownloadPdf(mocker):
    mocked_downloadPdf = mocker.patch("osd_slides.main.Downloader.downloadPdf")
    data = [
        """<tr><td valign="top"><img alt="[TXT]" src="/icons/text.gif"/>
        </td><td><a href="01-intro.html">01-intro.html</a></td>
        <td align="right">2023-03-04 14:38
        </td><td align="right"> 19K</td><td> </td></tr>""",
        """<tr><td valign="top"><img alt="[TXT]" src="/icons/text.gif"/>
        </td><td><a href="02-history.html">02-history.html</a></td>
        <td align="right">2023-03-04 14:38  </td>
        <td align="right"> 20K</td><td> </td></tr>""",
    ]

    soup = []
    for datum in data:
        soup.append(BeautifulSoup(datum, "html.parser"))
    # TODO repetitive work maybe have a class structure?
    mocker.patch("osd_slides.main.Downloader.readAvailableHtmlDocs", return_value=soup)  # noqa: E501
    downloader.searchAndDownloadPdf()
    #  check what is download called with assert_called_with
    calls = [call("01-intro.html"), call("02-history.html")]
    mocked_downloadPdf.assert_has_calls(calls)


# TODO unsuccessful test for searchAndDownloadPdf


@patch("builtins.print")
def test_showDownloadablePdf0(mock_print, mocker):
    #  successful read
    data = [
        """<tr><td valign="top"><img alt="[TXT]" src="/icons/text.gif"/></td><td><a href="01-intro.html">01-intro.html</a></td><td align="right">2023-03-04 14:38  </td><td align="right"> 19K</td><td> </td></tr>""",  # noqa: E501
        """<tr><td valign="top"><img alt="[TXT]" src="/icons/text.gif"/></td><td><a href="02-history.html">02-history.html</a></td><td align="right">2023-03-04 14:38  </td><td align="right"> 20K</td><td> </td></tr>""",  # noqa: E501
    ]
    soup = []
    for datum in data:
        soup.append(BeautifulSoup(datum, "html.parser"))

    mocker.patch("osd_slides.main.Downloader.readAvailableHtmlDocs", return_value=soup)  # noqa: E501
    downloader.showDownloadablePdf()

    assert mock_print.call_args_list[0].args == ("01-intro.html",)
    assert mock_print.call_args_list[1].args == ("02-history.html",)


# TODO unsuccessful test for downloadable


def test_readAvailableHtmlDocs(mocker):
    # TODO work on this
    # make requests.get send a weird
    # make soup have a predetermied value
    text = """
<html>
<head>
<title>Index of /~paine/4995/lectures</title>
</head>
<body>
<h1>Index of /~paine/4995/lectures</h1>
<table>
<tr><th valign="top"><img alt="[ICO]" src="/icons/blank.gif"/></th><th><a href="?C=N;O=D">Name</a></th><th><a href="?C=M;O=A">Last modified</a></th><th><a href="?C=S;O=A">Size</a></th><th><a href="?C=D;O=A">Description</a></th></tr> # noqa: E501
<tr><th colspan="5"><hr/></th></tr>
<tr><td valign="top"><img alt="[PARENTDIR]" src="/icons/back.gif"/></td><td><a href="/~paine/4995/">Parent Directory</a></td><td> </td><td align="right">  - </td><td> </td></tr> # noqa: E501
<tr><td valign="top"><img alt="[TXT]" src="/icons/text.gif"/></td><td><a href="01-intro.html">01-intro.html</a></td><td align="right">2023-03-05 13:21  </td><td align="right"> 19K</td><td> </td></tr> # noqa: E501
<tr><td valign="top"><img alt="[TXT]" src="/icons/text.gif"/></td><td><a href="02-history.html">02-history.html</a></td><td align="right">2023-03-05 13:21  </td><td align="right"> 20K</td><td> </td></tr> # noqa: E501
<tr><td valign="top"><img alt="[TXT]" src="/icons/text.gif"/></td><td><a href="03-git.html">03-git.html</a></td><td align="right">2023-02-14 22:46  </td><td align="right"> 25K</td><td> </td></tr> # noqa: E501
</table>
<address>Apache/2.4.29 (Ubuntu) Server at www.cs.columbia.edu Port 443</address> # noqa: E501
</body></html>"""
    expectedOutput = [
        '<tr><td valign="top"><img alt="[TXT]" src="/icons/text.gif"/></td><td><a href="01-intro.html">01-intro.html</a></td><td align="right">2023-03-05 13:21  </td><td align="right"> 19K</td><td> </td></tr>',  # noqa: E501
        '<tr><td valign="top"><img alt="[TXT]" src="/icons/text.gif"/></td><td><a href="02-history.html">02-history.html</a></td><td align="right">2023-03-05 13:21  </td><td align="right"> 20K</td><td> </td></tr>',  # noqa: E501
        '<tr><td valign="top"><img alt="[TXT]" src="/icons/text.gif"/></td><td><a href="03-git.html">03-git.html</a></td><td align="right">2023-02-14 22:46  </td><td align="right"> 25K</td><td> </td></tr>',  # noqa: E501
    ]  # noqa: E501
    soup = MagicMock(text=text)
    # soup["text"] = text
    mocker.patch("requests.get", return_value=soup)
    output = downloader.readAvailableHtmlDocs()
    for i in range(len(output)):
        output[i] = str(output[i])
    assert output == expectedOutput


# TODO unvalid URL testing needed


def test_downloadPdf():
    """
    It takes time as it needs to download a file.
    :return:
    """
    file = "01-intro.html"
    if os.path.isfile("./01-intro.pdf"):
        os.remove("./01-intro.pdf")
    downloader.downloadPdf(file)
    assert os.path.isfile("./01-intro.pdf")
    os.remove("./01-intro.pdf")


def test_search_init0():
    search = Search(url)
    assert search.text == []


def test_search_init1():
    search = Search(url, ["22-legal.html"])
    # TODO perhaps soup data can be inserted like above
    assert search.text[0][2][0] == (
        '## Major Legal Areas In the field of software development, three areas are important.'
    )


@patch("builtins.print")
def test_lookup0(mock_print):
    '''
    Testing lookup with no result
    :param mock_print:
    :return:
    '''
    search = Search(url)
    search.text = [defaultdict(dict)] * 1  # nothing inserted
    search.lookup("test")
    # check what is being printed
    assert mock_print.call_args_list[0].args == ("No slides were found",)


@patch("builtins.print")
def test_lookup1(mock_print):
    '''
    Testing lookup with no result
    :param mock_print:
    :return:
    '''
    search = Search(url)
    search.text = [
        dict(
            {
                1: {
                    0: "## Legal We'll break our discussion of legal issues into three categories",
                    1: 'Legal Fields Copyright Trademark Patent',
                    2: 'Project Copyright and Licenses',
                    3: 'Major Issues and Court Cases Oracle v. Google Cloud licensing / anticompetitive practices',
                }
            }
        )
    ]
    search.slides = ["slide1"]
    search.lookup("Legal")
    # check what is being printed
    assert mock_print.call_args_list[0].args == ("Slides that can be found are",)
    assert mock_print.call_args_list[1].args == ([(1, 0, 'slide1'), (1, 1, 'slide1')],)


def test_open0(mocker):
    # calls webopen with certain parameters!
    mocked_webbrowser_open = mocker.patch("webbrowser.open")
    search = Search(url)
    search.answers = [(1, 0, 'slide1'), (1, 1, 'slide1')]
    search.open()
    mocked_webbrowser_open.assert_called_once_with('https://www.cs.columbia.edu/~paine/4995/lectures/slide1#/1/0')


def test_open1(mocker):
    # calls webopen with certain parameters!
    mocked_webbrowser_open = mocker.patch("webbrowser.open")
    search = Search(url)
    search.answers = [(1, 0, 'slide1'), (1, 1, 'slide1')]
    search.open(1)
    mocked_webbrowser_open.assert_called_once_with('https://www.cs.columbia.edu/~paine/4995/lectures/slide1#/1/1')


def test_open2(mocker):
    '''
    Test with index value used
    :param mocker:
    :return:
    '''
    # calls webopen with certain parameters!
    mocked_webbrowser_open = mocker.patch("webbrowser.open")
    search = Search(url)
    search.answers = [(1, 0, 'slide1'), (1, 1, 'slide1')]

    with pytest.raises(IndexError) as e_info:
        search.open(2)


def test_open3():
    '''
    throws error when lookup is not performed before
    :return:
    '''
    search = Search(url)

    with pytest.raises(FileNotFoundError) as e_info:
        search.open()


# temp here


# INTEGRATION TESTS
@patch("builtins.print")
def test_print_downloadable_integration(mock_print):
    # mocked_searchAndDownloadPdf =
    # mocker.patch('osd_slides.main.searchAndDownloadPdf')
    testargs = ["prog", "show"]
    #  TODO improve the testing by mocking
    with patch.object(sys, "argv", testargs):
        downloader.fefmalsmc()
        assert mock_print.call_args_list[0].args == ("01-intro.html",)
        assert mock_print.call_args_list[1].args == ("02-history.html",)
        assert mock_print.call_args_list[2].args == ("03-git.html",)


# def test_download_integration():
#     """
#     Takes a long time as it needs to download good amount of files.
#     """
#     testargs = ["prog", "download"]
#     check = [
#         "01-intro.pdf",
#         "./02-history.pdf",
#         "./03-git.pdf",
#         "./03-projects.pdf",
#         "./04-agile-kanban.pdf",
#         "./04-git.pdf",
#         "./04-starting.pdf",
#         "./05-gh-issues-projects.pdf",
#         "./05-starting.pdf",
#         "./05-testing.pdf",
#         "./06-agile-kanban.pdf",
#         "./06-static.pdf",
#         "./06-testing-libraries.pdf",
#         "./07-docs.pdf",
#         "./07-docs2.pdf",
#         "./07-gh-issues-projects.pdf",
#         "./08-examples.pdf",
#         "./08-packaging-part-one.pdf",
#         "./08-release.pdf",
#         "./08-web.pdf",
#         "./09-examples.pdf",
#         "./09-release.pdf",
#         "./09-testing.pdf",
#         "./10-cicd.pdf",
#         "./10-problems.pdf",
#         "./10-static.pdf",
#         "./11-community.pdf",
#         "./11-problems2.pdf",
#         "./11-static.pdf",
#         "./11-testing-libraries.pdf",
#         "./12-econ.pdf",
#         "./12-testing-libraries.pdf",
#         "./13-legal.pdf",
#         "./13-packaging-part-two.pdf",
#         "./14-cicd-setup.pdf",
#         "./14-final.pdf",
#         "./15-release.pdf",
#     ]
#     for doc in check:
#         if os.path.isfile(doc):
#             os.remove(doc)
#
#     with patch.object(sys, "argv", testargs):
#         fefmalsmc()
#
#         for doc in check:
#             assert os.path.isfile(doc)
#             os.remove(doc)


@patch("builtins.print")
def test_search_integration(mock_print, mocker):
    '''
    Webdriver in the end is not called
    :param mocker:
    :return:
    '''
    mocked_webbrowser_open = mocker.patch("webbrowser.open")
    search = Search(url, ["22-legal.html"])
    search.lookup('Legal')
    search.open(2)
    mocked_webbrowser_open.assert_called_once_with(
        'https://www.cs.columbia.edu/~paine/4995/lectures/22-legal.html#/2/0'
    )



from main import Downloader

if __name__ == "__main__":
    # TODO dont know how package this python code into a
    # service run from terminal this will do for now
    downloader = Downloader()
    downloader.fefmalsmc()



import requests
from bs4 import BeautifulSoup
import sys
import subprocess

from osd_slides.search import Search

# TODO perhaps better to use properties file
#  using jproperties here instead of using a string.
url = "https://www.cs.columbia.edu/~paine/4995/lectures/"


class Downloader:
    def fefmalsmc(self):
        if len(sys.argv) >= 4:  # perhaps should be 3 now
            print("Too many arguments")
            # perhaps should throw an error or something

        elif len(sys.argv) < 2:
            print("Too few arguments")
        # length 2 acceptable
        else:
            if sys.argv[1] == "download":
                self.searchAndDownloadPdf()

            elif sys.argv[1] == "show":
                self.showDownloadablePdf()

            elif sys.argv[1] == "search":
                print("Which slides would you like to search in? Please provide slides with spaces between them: ")
                slides = input().split()
                print("Slides reading completed")
                # a = Search(url, [sys.argv[2]])
                a = Search(url, slides)
                print("What keyword would you like to look up: ")
                a.lookup(input())
                print("Which one would you like to open? please provide the number.")
                a.open(int(input()) - 1)
                # TODO change it to make it better

    # TODO be able to download individual files
    def searchAndDownloadPdf(self):
        '''Downloads all the class presentations.

        Returns:
            Null
        '''
        docs = self.readAvailableHtmlDocs()

        for doc in docs:
            try:
                file = doc.find("a")["href"]
                # file = doc
                self.downloadPdf(file)
            except TypeError:
                pass

        print("done")

    def showDownloadablePdf(self):
        '''Shows all the downloaddable pdf's.

        Returns:
            Null
        '''
        docs = self.readAvailableHtmlDocs()

        for doc in docs:
            print(doc.find("a")["href"])
            # print(doc)

    def readAvailableHtmlDocs(self):
        # TODO handle when response is not 200
        response = requests.get(url)
        soup = BeautifulSoup(response.text, "html.parser")
        table = soup.find("table")
        table_rows = table.find_all("tr")
        blacklist = {
            "example.html",
        }
        docs = []
        for row in table_rows:
            if row.find("a") and row.find("a")["href"].endswith(".html"):
                link = row.find("a")["href"]
                if link in blacklist:  # blacklisted
                    continue
                docs.append(row)
                # TODO we can change this to link to make it simpler

        return docs

    def downloadPdf(self, file):
        print("Downloading file: ", file)
        try:
            a = subprocess.run(
                ["npx", "decktape", "reveal", url + file + "#/", file[:-4] + "pdf"],  # noqa: E501
                capture_output=True,
                check=True,
            )
            print(a.stderr)
        except subprocess.CalledProcessError as exc:
            print("Standard error was {}".format(exc.stderr))

        # print(a.stdout)
        print("File ", file, " downloaded")



from bs4 import BeautifulSoup
import requests
from collections import defaultdict
import re
import webbrowser


class Search:
    # hash algorithms to library
    # Boyer Moore algorithm or KMP could be used
    # https://stackoverflow.com/questions/56418773/when-is-rabin-karp-more-effective-than-kmp-or-boyer-moore
    # or a custom page hash

    def __init__(self, url="https://www.cs.columbia.edu/~paine/4995/lectures/", slides=[]):
        '''
        Initializes search object. Prepares the slides to be searched for a keyword.
        :param url: Url for the class slides
        :param slides: List that contains slide names that needs to be searched. Slide names can be found by using
        Downloader.showDownloadablePdf()
        '''
        self.text = [defaultdict(dict) for _ in range(len(slides))]

        self.answers = []
        self.slides = slides
        self.url = url  # TODO this should be read from properties file

        for z, slide in enumerate(slides):
            response = requests.get(url + slide)
            soup = BeautifulSoup(response.text, "html.parser")
            table = soup.find("div", class_="slides")
            sections = table.find_all("section", attrs={'data-markdown': True})

            for idx, section in enumerate(sections):
                if "data-separator" in section.attrs:  # has multiple pages in one x axis value
                    seperator = section["data-separator"]
                    non_seperated_text = section.contents[1].text  # current all the stuff

                    seperated_text = non_seperated_text.split(seperator)
                    for yindx, individual_text in enumerate(seperated_text):
                        temp = re.sub("(<!--.*?-->)", "", individual_text, flags=re.DOTALL)  # removes html comments
                        self.text[z][idx][yindx] = re.sub(r"\s+", " ", temp).strip()  # removes empty spaces
                else:  # only has one page in that
                    temp = re.sub(
                        "(<!--.*?-->)", "", section.contents[1].text, flags=re.DOTALL
                    )  # removes html comments
                    self.text[z][idx][0] = re.sub(r"\s+", " ", temp).strip()  # removes empty spaces
            # https://stackoverflow.com/questions/761824/python-how-to-convert-markdown-formatted-text-to-text

    def lookup(self, keyword):
        '''
        Looks up the keyword in slides.

        keyword: Keyword is searched through the slides.

        Returns:
            Null
        '''

        # it will look up the word in the dict and come up with possible answers
        self.answers = []  # resetted
        for i in range(len(self.text)):
            for idx, outer in self.text[i].items():
                for yidx, inner in outer.items():
                    if keyword in inner:
                        self.answers.append(tuple([idx, yidx, self.slides[i]]))
        if not self.answers:
            print("No slides were found")
            return
        else:
            print("Slides that can be found are")
            print(self.answers)
            return
        #
        # TODO perhaps make all lowercase

    # TODO show lookup results only without writing over

    def open(self, idx=0):
        '''
        lookup() must be called fist.
        Opens the exact slide were the keyword was found using index.

        idx: Index to decide which exact slide to open.

        Returns:
                Null
        '''
        if not self.answers:
            raise FileNotFoundError("Lookup not performed in order to open")
        if idx >= len(self.answers):  # idx too big
            raise IndexError("index out of bounds")
        webbrowser.open(
            "{}{}#/{}/{}".format(self.url, self.answers[idx][2], self.answers[idx][0], self.answers[idx][1])
        )
